#ifndef SOLVER_H
#define SOLVER_H

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <vector>
#include <list>
#include <algorithm>
using namespace std;
#define EPS 1e-15

typedef vector<int> VI;
typedef vector<VI> VVI;
typedef list<pair<double,double> > LPDD;

inline double convertFrom01(double v) { return 1 - 2*v; }
inline double convertTo01Real(double v) { return (1-v) / 2; }
inline char convertToLabel(double v) { return (v<0 || (v==0 && (rand()&1))) ? '1' : '0'; }
inline int sign(double v) { return v>0 ? 1 : (v<0 ? -1 : 0); }

class Solver {
protected:
	int n, l;
	VVI label2code, code2label;
	double *val, *x, *product;
	inline double getProductExcept(int pid, int xid) {
		if(fabs(x[xid]) >= EPS) return product[pid] / x[xid];
		else {
			double w = 1;
			for(VI::iterator itr = code2label[pid].begin(); itr != code2label[pid].end(); ++itr) {
				if(*itr != xid) w *= x[*itr];
			}
			return w;
		}
	}
public:
	Solver(int _n, int _l):n(_n), l(_l), label2code(_n), code2label(_l) { val = x = product = NULL; }
	virtual void solveIter(char *fixed = NULL) = 0;
	virtual double calculateLoss() = 0;
	virtual VVI& getLabel2Code() { return label2code; }
	virtual VVI& getCode2Label() { return code2label; }
	virtual void solveSetup(double *_val, double *_x, char *fixed = NULL, double *init = NULL) {
		x = _x;
		val = (double*)malloc(sizeof(double)*l);
		memcpy(val, _val, sizeof(double)*l);
		product = (double*)malloc(sizeof(double)*l);
		for(int i = 0; i < n; ++i) {
			if(fixed != NULL && fixed[i] >= 0) x[i] = convertFrom01(fixed[i]);
			else if(init != NULL) x[i] = init[i];
			else x[i] = rand()*2.0/RAND_MAX - 1;
		}
		for(int i = 0; i < l; ++i) {
			product[i] = 1.0;
			for(int j = 0; j < int(code2label[i].size()); ++j)
				product[i] *= x[code2label[i][j]];
		}
	}
	virtual void solveFinal() {
		free(val);
		free(product);
		val = x = product = NULL;
	}
	virtual void updateX(int id, double v) {
		v = max(-1.0, min(v, 1.0));
		if(fabs(x[id]) < EPS || fabs(v) < EPS) {
			x[id] = v;
			for(VI::iterator itr = label2code[id].begin(); itr != label2code[id].end(); ++itr) {
				product[*itr] = 1;
				for(VI::iterator jtr = code2label[*itr].begin(); jtr != code2label[*itr].end(); ++jtr)
					product[*itr] *= x[*jtr];
			}
		} else {
			double ratio = v / x[id];
			for(VI::iterator itr = label2code[id].begin(); itr != label2code[id].end(); ++itr) {
				product[*itr] *= ratio;
			}
			x[id] = v;
		}
	}
};

class SolverL2 : public Solver {
public:
	SolverL2(int _n, int _l):Solver(_n, _l) {}
	virtual void solveIter(char *fixed = NULL) {
		for(int i = 0; i < n; ++i) {
			if(fixed != NULL && fixed[i] >= 0) continue;
			double sumVQ = 0, sumQQ = 0;
			for(VI::iterator itr = label2code[i].begin(); itr != label2code[i].end(); ++itr) {
				sumVQ += val[*itr] * product[*itr];
				sumQQ += product[*itr] * product[*itr];
			}
			double ratio = sumVQ / sumQQ;
			updateX(i, x[i]*ratio);
		}
	}
	virtual double calculateLoss() {
		double r = 0;
		for(int i = 0; i < l; ++i) r += (product[i]-val[i]) * (product[i]-val[i]);
		return r;
	}
};

class SolverL1 : public Solver {
protected:
	double findMedian(LPDD &wy, double w) {
		assert(!isnan(w));
		double pivot = wy.front().second, wlt = 0, wgt = 0, weq = 0;
		for(LPDD::iterator itr = wy.begin(); itr != wy.end(); ++itr) {
			if(itr->second < pivot) wlt += itr->first;
			else if(itr->second == pivot) weq += itr->first;
			else wgt += itr->first;
		}
		if(wlt < w && w < wlt+weq) return pivot;
		else if(w <= wlt) {
			for(LPDD::iterator itr = wy.begin(); itr != wy.end();) {
				if(itr->second >= pivot) itr = wy.erase(itr);
				else ++itr;
			}
			double p2 = (wy.size()>0) ? findMedian(wy, w) : pivot;
			return (w==wlt) ? (p2+pivot/2) : p2;
		} else {	// w >= wlt+weq
			for(LPDD::iterator itr = wy.begin(); itr != wy.end();) {
				if(itr->second <= pivot) itr = wy.erase(itr);
				else ++itr;
			}
			double p2 = (wy.size()>0) ? findMedian(wy, w-(wlt+weq)) : pivot;
			return (w==wlt+weq) ? (p2+pivot/2) : p2;
		}
	}
public:
	SolverL1(int _n, int _l):Solver(_n, _l) {}
	virtual void solveIter(char *fixed = NULL) {
		LPDD wy(l);
		for(int i = 0; i < n; ++i) {
			if(fixed != NULL && fixed[i] >= 0) continue;
			wy.clear();
			double sumW = 0;
			random_shuffle(label2code[i].begin(), label2code[i].end());
			for(VI::iterator itr = label2code[i].begin(); itr != label2code[i].end(); ++itr) {
				double w = getProductExcept(*itr, i);
				if(fabs(w) >= EPS) {
					sumW += fabs(w);
					wy.push_back(make_pair(fabs(w), val[*itr]/w));
				}
			}
			updateX(i, findMedian(wy, sumW/2));
		}
	}
	virtual double calculateLoss() {
		double r = 0;
		for(int i = 0; i < l; ++i) r += fabs(product[i]-val[i]);
		return r;
	}
};

class SolverL1Osr : public Solver {
private:
	struct Event {
		double pos, slope;
		Event() {}
		Event(double p, double s):pos(p),slope(s) {}
		bool operator<(const struct Event &a) const {
			return pos < a.pos;
		}
	};
public:
	SolverL1Osr(int _n, int _l):Solver(_n, _l) {}
	virtual void solveIter(char *fixed = NULL) {
		for(int i = 0; i < n; ++i) {
			if(fixed != NULL && fixed[i] >= 0) continue;
			list<Event> evq(label2code[i].size()*2);
			double start = 0, slope = 0;
			for(VI::iterator itr = label2code[i].begin(); itr != label2code[i].end(); ++itr) {
				double w = getProductExcept(*itr, i), y = val[*itr];
				double absw = fabs(w);
				if(absw < EPS) continue;
				double xx = y / w;
				if(sign(w)*sign(y) < 0) {
					evq.push_back(Event(max(-1.0,xx), absw));
					evq.push_back(Event(1, -absw));
				} else if(sign(w)*sign(y) > 0) {
					evq.push_back(Event(min(xx,1.0), absw));
					start += (xx+1) * absw;
					slope += -absw;
				}
			}
			evq.sort();
			double bestX = -1, bestV = start, nowX = -1, nowV = start;
			while(!evq.empty()) {
				double nxtX = evq.front().pos;
				double nxtV = nowV + (nxtX-nowX)*slope;
				if(nxtV < bestV) bestV = start, bestX = nxtX;
				else if(nxtV == bestV && nowV == nxtV) bestX = (nowX+nxtX) / 2;
				nowX = nxtX, nowV = nxtV;
				while(!evq.empty() && evq.front().pos == nowX) {
					slope += evq.front().slope;
					evq.pop_front();
				}
			}
			if(nowV == bestV) bestX = (nowV+1) / 2;
			updateX(i, bestX);
		}
	}
	virtual double calculateLoss() {
		double r = 0;
		for(int i = 0; i < l; ++i) r += max((val[i]-product[i])*sign(val[i]), 0.0);
		return r;
	}
};

double solveSquareLoss(int n, int l, VVI label2code, VVI code2label, double *val, double *x,
		char *fixed = NULL, double *init = NULL, int iter = 50) {
	double *prod = (double*)malloc(sizeof(double)*l);
	for(int i = 0; i < n; ++i) {
		if(fixed != NULL && fixed[i] >= 0) x[i] = convertFrom01(fixed[i]);
		else if(init != NULL) x[i] = init[i];
		else x[i] = rand()/double(RAND_MAX) - 0.5;
	}
	for(int i = 0; i < l; ++i) {
		prod[i] = 1.0;
		for(int j = 0; j < int(code2label[i].size()); ++j)
			prod[i] *= x[code2label[i][j]];
	}
	for(int i = 0; i < iter; ++i) {
		for(int j = 0; j < n; ++j) {
			if(fixed != NULL && fixed[j] >= 0) continue;
			double sumVQ = 0.0, sumQQ = 0.0;
			for(int k = 0; k < int(label2code[j].size()); ++k) {
				int c = label2code[j][k];
				sumVQ += val[c] * prod[c];
				sumQQ += prod[c] * prod[c];
			}
			double ratio = sumVQ / sumQQ;
			double nx = x[j] * ratio;
			nx = (nx<-1) ? -1 : ((nx>1) ? 1 : nx);
			ratio = nx / x[j];
			for(int k = 0; k < int(label2code[j].size()); ++k)
				prod[label2code[j][k]] *= ratio;
			x[j] = nx;
		}
	}
	double s = 0.0;
	for(int i = 0; i < l; ++i) s += (prod[i]-val[i]) * (prod[i]-val[i]);
	free(prod);
	return s;
}

#endif
