function [P] = gmmPredict(mdl, X)
% Predict probability of X in each class: normalize(likelihood*prior)
% mdl = {'ys'=[y], 'prior'={y=>p}, 'gmms'={y=>{'col'=[],'gmd'=GMDISTRIBUTION}}}
	if ischar(mdl)
		mdl = load('-mat', mdl);
	end

	ys = mdl.ys;
	if size(ys, 2) == 1
		P = repmat([1.0], size(X,1), 1);
		return
	end
	P = [];
	prior = [];
	gmms = {mdl.gmms};
	pr = {mdl.prior};
	for y = ys
		gmm = gmms{y+1};
		prior = [prior pr{y+1}];
		col = gmm.col;
		x = X(:, col);
		P = [P gmmLogPdf(gmm.gmd, x)];
	end
	P = P + repmat(log(prior), size(P,1), 1);
	px = logsumexp(P, 2);
	P = exp(P - repmat(px, 1, size(P,2)));
end

function [logP] = gmmLogPdf(gmd, X)
% The original pdf in gmdistribution class may overflow easily.
	covNames = {'diagonal', 'full'};
	CovType = find(strncmpi(gmd.CovType,covNames,length(gmd.CovType)));
	loglh = wdensity(X, gmd.mu, gmd.Sigma, gmd.PComponents, gmd.SharedCov, CovType);
	logP = logsumexp(loglh, 2);
end

function s = logsumexp(x, dim)
% Compute log(sum(exp(x),dim)) while avoiding numerical underflow.
%   By default dim = 1 (columns).
% Written by Michael Chen (sth4nth@gmail.com).
	if nargin == 1, 
	% Determine which dimension sum will use
		dim = find(size(x)~=1,1);
		if isempty(dim), dim = 1; end
	end

	% subtract the largest in each column
	y = max(x,[],dim);
	x = bsxfun(@minus,x,y);
	s = y + log(sum(exp(x),dim));
	i = find(~isfinite(y));
	if ~isempty(i)
		s(i) = y(i);
	end
end

% ==================== Internal code in MATLAB ====================

function   [log_lh,mahalaD]=wdensity(X, mu, Sigma, p, sharedCov, CovType)
%WDENSITY Weighted conditional density and mahalanobis distance.
%   LOG_LH = WDENSITY(...) returns log of component conditional density
%   (weighted by the component probability) of X. LOG_LH is a N-by-K matrix
%   LOG_LH, where K is the number of Gaussian components. LOG_LH(I,J) is
%   log (Pr(point I|component J) * Prob( component J))
%
%   [LOG_LH, MAHALAD]=WDENSITY(...) returns the Mahalanobis distance in
%   the N-by-K matrix MAHALAD. MAHALAD(I,J) is the Mahalanobis distance of
%   point I from the mean of component J.

%   Copyright 2007 The MathWorks, Inc.
%   $Revision: 1.1.8.2 $  $Date: 2011/05/09 01:28:14 $

    log_prior = log(p);
    [n,d]=size(X);
    k=size(mu,1);
    log_lh = zeros(n,k);
    mahalaD = zeros(n,k);
    logDetSigma = -Inf;
    for j = 1:k
        if sharedCov
            if j == 1
                if CovType == 2 % full covariance
                    [L,f] = chol(Sigma);
                    diagL = diag(L);
                    if (f ~= 0)|| any(abs(diagL) < eps(max(abs(diagL)))*size(L,1))
                        error(message('stats:gmdistribution:wdensity:IllCondCov'));
                    end
                    logDetSigma = 2*sum(log(diagL));
                else %diagonal
                    L = sqrt(Sigma);
                    if  any(L < eps( max(L))*d)
                          error(message('stats:gmdistribution:wdensity:IllCondCov'));
                    end
                    logDetSigma = sum( log(Sigma) );
                end
            end
        else %different covariance
            if CovType == 2 %full covariacne
                % compute the log determinant of covariance
                [L,f] = chol(Sigma(:,:,j) );
                diagL = diag(L);
                if (f ~= 0) || any(abs(diagL) < eps(max(abs(diagL)))*size(L,1))
                     error(message('stats:gmdistribution:wdensity:IllCondCov'));
                end
                logDetSigma = 2*sum(log(diagL));
            else %diagonal covariance
                L = sqrt(Sigma(:,:,j)); % a vector
                if  any(L < eps(max(L))*d)
                     error(message('stats:gmdistribution:wdensity:IllCondCov'));
                end
                logDetSigma = sum( log(Sigma(:,:,j)) );

            end
        end

        Xcentered = bsxfun(@minus, X, mu(j,:));
       
        if CovType == 2
            xRinv = Xcentered /L ;
        else
            xRinv = bsxfun(@times,Xcentered , (1./ L));
        end
        mahalaD(:,j) = sum(xRinv.^2, 2);

        log_lh(:,j) = -0.5 * mahalaD(:,j) +...
            (-0.5 *logDetSigma + log_prior(j)) - d*log(2*pi)/2;
        %get the loglikelihood for each point with each component
        %log_lh is a N by K matrix, log_lh(i,j) is log \alpha_j(x_i|\theta_j)
    end
end
   
